package algorithm;

import org.junit.Test;

import java.util.*;

/**
 * @author wmy3969
 * @version 1.0
 * @date 2022/5/9 22:42
 * @Description
 */
public class AlgorithmExample {
    @Test
    public void testRecursion() {
        int width = 10;
        int height = 10;
        int[][] map = new int[width][height];
        this.setBarrier(map);// 设置障碍
        this.printMap(map, width, height);// 打印map
        final boolean result = findWay(map, 8, 1);
        this.printMap(map, width, height);// 打印map
    }

    /**
     * 找路
     *
     * @param map 迷宫地图
     * @param i   尝试点i
     * @param j   尝试点j
     * @return 此点往前走成败
     */
    private boolean findWay(int[][] map, int i, int j) {
        /*printMap(map, 10, 10);*/
        if (map[1][1] == 1) {// 设置出口
            return true;
        } else {
            if (map[i][j] == 0) {
                map[i][j] = 1;
                if (findWay(map, i, j + 1)) {// 右
                    return true;
                } else if (findWay(map, i + 1, j)) {// 下
                    return true;
                } else if (findWay(map, i - 1, j)) {// 上
                    return true;
                } else if (findWay(map, i, j - 1)) {// 左
                    return true;
                } else {
                    map[i][j] = 4;
                    return false;
                }
            } else {
                return false;
            }
        }
    }

    /**
     * 打印迷宫地图
     *
     * @param map    迷宫地图
     * @param width  宽度
     * @param height 高度
     */
    private void printMap(int[][] map, int width, int height) {
        for (int i = 0; i < height; i++) {
            for (int j = 0; j < width; j++) {
                System.out.print(map[i][j] + "\t");
            }
            System.out.println();
            System.out.println();
        }
        System.out.println();
    }

    /**
     * 给迷宫设置障碍
     *
     * @param map 迷宫地图
     */
    private void setBarrier(int[][] map) {
        for (int i = 0; i < map.length; i++) {
            map[0][i] = 8;
            map[map.length - 1][i] = 8;
            map[i][0] = 8;
            map[i][map[0].length - 1] = 8;
        }
        map[6][1] = 8;
        map[6][2] = 8;
        map[6][3] = 8;
        map[6][4] = 8;
        map[6][5] = 8;
        map[3][3] = 8;
        map[3][4] = 8;
        map[3][5] = 8;
        map[3][6] = 8;
        map[3][7] = 8;
        map[3][8] = 8;
        map[3][9] = 8;
        map[5][7] = 8;
        map[6][7] = 8;
        map[7][7] = 8;
        map[1][4] = 8;
    }

    @Test
    public void testBacktracking() {
        int[] map = new int[8];
        setPosition(map, 0);
        System.out.println("解法种数: " + ansCnt);
        System.out.println("总计次数: " + allCnt);
    }

    int ansCnt = 0;
    int allCnt = 0;

    private void setPosition(int[] map, int n) {
        if (Objects.equals(n, 8)) {
            System.out.println(Arrays.toString(map));
            ansCnt++;
            return;
        }
        for (int i = 0; i < map.length; i++) {
            map[n] = i;
            if (check(map, n)) {
                setPosition(map, n + 1);
            }
            allCnt++;
        }
    }

    /**
     * 检查是否冲突
     *
     * @param map 棋盘
     * @param n   第n个皇后
     * @return 是否冲突
     */
    private boolean check(int[] map, int n) {
        for (int i = 0; i < n; i++) {
            // Objects.equals(map[i], map[n])->表示在同一列
            // Objects.equals(Math.abs(map[i] - map[n]), Math.abs(i - n))->表示在同一斜线
            if (Objects.equals(map[i], map[n]) || Objects.equals(Math.abs(map[i] - map[n]), Math.abs(i - n))) {
                return false;
            }
        }
        return true;
    }

    /**
     * 入口方法
     */
    @Test
    public void main() {
        init();
        makeMaze();
        //printMaze();
        findMazeWay(maze, 1, 1);
        printMaze();
    }

    int len = 41; // 迷宫长度
    int wid = 41; // 迷宫宽度
    char barrier = '■'; // 代表障碍
    char way = '○'; // 代表路
    char[][] maze; // 迷宫
    boolean[][] visit; // ⽤来标记某⼀格是否被访问过
    MazeNode start = new MazeNode(1, 1); // 开始节点
    MazeNode exit = new MazeNode(len - 1, wid - 1); // 出⼝节点
    MazeNode cur; //当前格
    MazeNode next; //下⼀格
    Stack<MazeNode> path = new Stack<MazeNode>(); //表⽰路径的栈
    int[][] adj = {{0, 2}, {0, -2}, {2, 0}, {-2, 0}}; //⽤来计算邻接格

    /**
     * 迷宫找路策略
     *
     * @param maze 迷宫地图
     * @param x    待找点x坐标
     * @param y    待找点y坐标
     * @return 此点能否继续朝前走
     */
    private boolean findMazeWay(char[][] maze, int x, int y) {
        if (x > len - 1 || y > wid - 1) {
            return false;
        }
        if (maze[exit.x - 1][exit.y - 1] == '1') {// 设置出口
            maze[exit.x - 1][exit.y - 1] = '√';
            return true;
        } else {
            if (maze[x][y] == '○') {
                maze[x][y] = '1';
                if (findMazeWay(maze, x, y + 1)) {// 右
                    if (maze[x][y] == '1')
                        maze[x][y] = '→';
                    return true;
                } else if (findMazeWay(maze, x + 1, y)) {// 下
                    if (maze[x][y] == '1')
                        maze[x][y] = '↓';
                    return true;
                } else if (findMazeWay(maze, x - 1, y)) {// 上
                    if (maze[x][y] == '1')
                        maze[x][y] = '↑';
                    return true;
                } else if (findMazeWay(maze, x, y - 1)) {// 左
                    if (maze[x][y] == '1')
                        maze[x][y] = '←';
                    return true;
                } else {
                    maze[x][y] = '×';
                    return false;
                }
            } else {
                return false;
            }
        }
    }

    /**
     * 迷宫的格子类
     */
    public static class MazeNode {
        int x, y;

        public MazeNode() {
        }

        public MazeNode(int x, int y) {
            this.x = x;
            this.y = y;
        }

        public String toString() {
            return "Node [x=" + x + ", y=" + y + "]";
        }
    }

    /**
     * 初始化，初始化迷宫参数
     */
    public void init() {
        maze = new char[len][wid];
        visit = new boolean[len][wid];
        for (int i = 0; i < len; i++) {
            for (int j = 0; j < wid; j++) {
                maze[i][j] = barrier;
                visit[i][j] = false;
            }
        }
        visit[start.x][start.y] = true;
        maze[start.x][start.y] = way;
        cur = start; //将当前格标记为开始格
    }

    /**
     * 打印迷宫地图
     */
    public void printMaze() {
        for (int i = 0; i < len; i++) {
            for (int j = 0; j < wid; j++) {
                System.out.print(maze[i][j] + " ");
                /*if (maze[i][j] == '○') {
                    System.err.print(maze[i][j] + " ");
                } else {
                    System.out.print(maze[i][j] + " ");
                }
                try {
                    Thread.sleep(100);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }*/
            }
            System.out.println();
        }
        System.out.println();
    }

    /**
     * 开始制作迷宫
     */
    public void makeMaze() {
        path.push(cur); //将当前格压⼊栈
        while (!path.empty()) {
            MazeNode[] adjs = notVisitedAdj(cur);//寻找未被访问的邻接格
            if (adjs.length == 0) {
                cur = path.pop();//如果该格⼦没有可访问的邻接格，则跳回上⼀个格⼦
                continue;
            }
            next = adjs[new Random().nextInt(adjs.length)]; //随机选取⼀个邻接格
            int x = next.x;
            int y = next.y;
            //如果该节点被访问过，则回到上⼀步继续寻找
            if (visit[x][y]) {
                cur = path.pop();
            } else {//否则将当前格压⼊栈，标记当前格为已访问，并且在迷宫地图上移除障碍物
                path.push(next);
                visit[x][y] = true;
                maze[x][y] = way;
                maze[(cur.x + x) / 2][(cur.y + y) / 2] = way; //移除当前格与下⼀个之间的墙壁
                cur = next;//当前格等于下⼀格
            }
        }
    }

    /**
     * 判断节点是否都被访问
     *
     * @param ns
     * @return
     */
    boolean allVisited(MazeNode[] ns) {
        return Arrays.stream(ns).allMatch(n -> visit[n.x][n.y]);
    }

    /**
     * 寻找可访问的邻接格
     *
     * @param mazeNode
     * @return
     */
    public MazeNode[] notVisitedAdj(MazeNode mazeNode) {
        List<MazeNode> list = new ArrayList<MazeNode>();
        for (int[] ints : adj) {
            int x = mazeNode.x + ints[0];
            int y = mazeNode.y + ints[1];
            if (x >= 0 && x < len && y >= 0 && y < wid) {
                if (!visit[x][y])
                    list.add(new MazeNode(x, y));
            }
        }
        MazeNode[] a = new MazeNode[list.size()];
        for (int i = 0; i < list.size(); i++) {
            a[i] = list.get(i);
        }
        return a;
    }

    @Test
    public void horseChessboard() {
        int[][] chessboard = new int[8][8];
        int[][] step = {{-1, -2}, {1, -2}, {-2, -1}, {2, -1}, {-2, 1}, {2, 1}, {-1, 2}, {1, 2}};
        int[] start = {5, 5};
        for (int i = 0; i < 8; i++) {
            for (int j = 0; j < 8; j++) {
                chessboard[i][j] = 0;
            }
        }
        nextStep(chessboard, step, start[0], start[1]);
        printChessboard(chessboard);
    }

    public boolean allVis(int[][] chessboard) {
        for (int i = 0; i < 8; i++) {
            for (int j = 0; j < 8; j++) {
                if (Objects.equals(chessboard[i][j], 0)) {
                    return false;
                }
            }
        }
        return true;
    }

    public boolean nextStep(int[][] chessboard, int[][] step, int x, int y) {
        if (allVis(chessboard)) {
            /*printChessboard(chessboard);*/
            return true;
        }
        if (chessboard[x][y] == 0) {
            chessboard[x][y] = 1;
            for (int i = 0; i < step.length; i++) {
                int nextX = x + step[i][1];
                int nextY = y + step[i][0];
                if (nextX >= 0 && nextY >= 0 && nextX < 8 && nextY < 8 && chessboard[nextX][nextY] == 0)
                    if (!nextStep(chessboard, step, nextX, nextY))
                        chessboard[x][y] = 1;
            }
            chessboard[x][y] = 0;
            return false;
        } else if (chessboard[x][y] == 1) {
            return false;
        }
        return true;
    }

    public void printChessboard(int[][] chessboard) {
        for (int i = 0; i < 8; i++) {
            for (int j = 0; j < 8; j++) {
                System.out.print(chessboard[i][j] + "\t");
            }
            System.out.println();
            System.out.println();
        }
    }


    /**
     * 棋盘的列
     */
    private int X;
    /**
     * 棋盘的行
     */
    private int Y;
    /**
     * 标记棋盘各个位置是否被访问过
     */
    private boolean[] visited;
    /**
     * 标记棋盘所有位置是否都被访问完
     */
    private boolean finished;

    @Test
    public void test() {
        // 棋盘的行和列都为8
        X = 12;
        Y = 12;
        // 骑士初始位置行和列都为1
        int row = 5;
        int column = 5;
        // 创建棋盘
        int[][] chessBoard = new int[X][Y];
        // 初始化棋盘所有位置都未被访问
        visited = new boolean[X * Y];

        // 测试一下耗时
        long start = System.currentTimeMillis();
        traversalChessBoard(chessBoard, row - 1, column - 1, 1);
        long end = System.currentTimeMillis();
        System.out.println("共耗时：" + (end - start) + "毫秒");

        // 输出棋盘的最后情况
        System.out.println("骑士的周游路径为：");
        for (int[] rows : chessBoard) {
            for (int step : rows) {
                System.out.print(step + "\t");
            }
            System.out.println();
            System.out.println();
        }
    }

    /**
     * 骑士周游棋盘算法
     *
     * @param chessBoard 棋盘
     * @param row        骑士当前位置的行
     * @param column     骑士当前位置的列
     * @param step       第几步，初始位置是第1步
     */
    public void traversalChessBoard(int[][] chessBoard, int row, int column, int step) {
        // 标记棋盘对应的位置为行走的第几步
        chessBoard[row][column] = step;
        // 标记棋盘对应位置已被访问过
        visited[row * X + column] = true;
        // 获取下一步可走位置的集合
        ArrayList<Point> ps = next(new Point(column, row));
        // 贪心算法优化：对ps排序，排序依据为：ps的下一步可选择数目，做非递减排序
        sort(ps);

        // 遍历ps
        while (!ps.isEmpty()) {
            // 取出下一个可走的位置
            Point p = ps.remove(0);
            // 判断该点是否已经访问过
            if (!visited[p.y * X + p.x]) {
                // 递归
                traversalChessBoard(chessBoard, p.y, p.x, step + 1);
            }
        }

        // 判断马儿是否完成任务
        if (step < X * Y && !finished) {
            // 如果没有完成，将整个棋盘置为0
            chessBoard[row][column] = 0;
            visited[row * X + column] = false;
        } else {
            finished = true;
        }
    }

    /**
     * 根据当前位置（Point对象），计算骑士还能走哪些位置（Point），并放入到ArrayList集合中，最多有8个位置
     *
     * @param curPoint 当前位置
     * @return 下一步还能走的位置集合
     */
    public ArrayList<Point> next(Point curPoint) {
        // Java的Point类，可表示坐标中的点
        ArrayList<Point> ps = new ArrayList<>();
        // 创建点对象
        Point p1 = new Point();

        // 判断骑士能不能走5这个位置
        p1.x = curPoint.x - 2;
        p1.y = curPoint.y - 1;
        if (p1.x >= 0 && p1.y >= 0) {
            ps.add(new Point(p1));
        }

        // 判断骑士能不能走6这个位置
        p1.x = curPoint.x - 1;
        p1.y = curPoint.y - 2;
        if (p1.x >= 0 && p1.y >= 0) {
            ps.add(new Point(p1));
        }

        // 判断骑士能不能走7这个位置
        p1.x = curPoint.x + 1;
        p1.y = curPoint.y - 2;
        if (p1.x < X && p1.y >= 0) {
            ps.add(new Point(p1));
        }

        // 判断骑士能不能走0这个位置
        p1.x = curPoint.x + 2;
        p1.y = curPoint.y - 1;
        if (p1.x < X && p1.y >= 0) {
            ps.add(new Point(p1));
        }

        // 判断骑士能不能走1这个位置
        p1.x = curPoint.x + 2;
        p1.y = curPoint.y + 1;
        if (p1.x < X && p1.y < Y) {
            ps.add(new Point(p1));
        }

        // 判断骑士能不能走2这个位置
        p1.x = curPoint.x + 1;
        p1.y = curPoint.y + 2;
        if (p1.x < X && p1.y < Y) {
            ps.add(new Point(p1));
        }

        // 判断骑士能不能走3这个位置
        p1.x = curPoint.x - 1;
        p1.y = curPoint.y + 2;
        if (p1.x >= 0 && p1.y < Y) {
            ps.add(new Point(p1));
        }

        // 判断骑士能不能走4这个位置
        p1.x = curPoint.x - 2;
        p1.y = curPoint.y + 1;
        if (p1.x >= 0 && p1.y < Y) {
            ps.add(new Point(p1));
        }

        return ps;
    }

    /**
     * 对当前这一步的所有的下一步的可选择位置，做非递减排序，每次都选择下一步的可走下一步最少的位置，减少回溯的次数
     * 1,2,3,4,5,.        :递增排列
     * 9,8,7,6,5,.        :递减排列
     * 1,2,3,3,4,5,8,8,.  :非递减排列
     * 9,8,7,7,6,5,5,2,1,.:非递增排列
     */
    public void sort(ArrayList<Point> ps) {
        ps.sort((o1, o2) -> {
            // 获取o1和o2的下一步的所有位置个数
            int count1 = next(o1).size();
            int count2 = next(o2).size();

            return Integer.compare(count1, count2);
        });
    }

    public static class Point {
        int x, y;

        public Point() {
        }

        public Point(int x, int y) {
            this.x = x;
            this.y = y;
        }

        public Point(Point point) {
            this.x = point.x;
            this.y = point.y;
        }
    }

}
